# -*- coding: utf-8 -*-

import pytest
import asyncio
import logging
import aiochrome
import functools


logging.basicConfig(level=logging.INFO)
logging.getLogger("urllib3").setLevel(logging.WARNING)


async def close_all_tabs(browser):
    tabs = await browser.list_tab()
    if len(tabs) == 0:
        return

    logging.debug("[*] recycle")
    for tab in tabs:
        await browser.close_tab(tab)

    await asyncio.sleep(1)
    assert len(browser.list_tab()) == 0


@pytest.mark.asyncio
async def new_multi_tabs(browser, n):
    tabs = []
    for i in range(n):
        tabs.append(await browser.new_tab())

    return tabs


@pytest.mark.asyncio
async def test_normal_callmethod():
    browser = aiochrome.Browser()
    tabs = await new_multi_tabs(browser, 10)

    for tab in tabs:
        await tab.start()
        result = await tab.Page.navigate(url="http://www.fatezero.org")
        assert result['frameId']

    await asyncio.sleep(3)

    for tab in tabs:
        result = await tab.Runtime.evaluate(expression="document.domain")
        assert result['result']['type'] == 'string'
        assert result['result']['value'] == 'www.fatezero.org'
        await tab.stop()


@pytest.mark.asyncio
async def test_set_event_listener():
    browser = aiochrome.Browser()
    tabs = await new_multi_tabs(browser, 10)

    async def request_will_be_sent(tab, **kwargs):
        await tab.stop()

    for tab in tabs:
        await tab.start()
        tab.Network.requestWillBeSent = functools.partial(request_will_be_sent, tab)
        await tab.Network.enable()
        try:
            await tab.Page.navigate(url="chrome://newtab/")
        except aiochrome.UserAbortException:
            pass

    for tab in tabs:
        if not await tab.wait(timeout=5):
            assert False, "never get here"
        await tab.stop()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import asyncio
import aiohttp

from .tab import Tab


__all__ = ["Browser"]


class Browser:
    _all_tabs = {}

    def __init__(self, url="http://127.0.0.1:9222", loop=None):
        self.dev_url = url
        self.loop = loop or asyncio.get_event_loop()

        if self.dev_url not in self._all_tabs:
            self._tabs = self._all_tabs[self.dev_url] = {}
        else:
            self._tabs = self._all_tabs[self.dev_url]

        self.session = aiohttp.ClientSession(loop=self.loop)

    async def new_tab(self, url=None, timeout=None):
        url = url or ''
        async with self.session.get("{}/json/new?{}".format(self.dev_url, url), timeout=timeout) as rp:
            tab_json = await rp.json()
            tab = Tab(**tab_json)
            self._tabs[tab.id] = tab
            return tab

    async def list_tab(self, timeout=None):
        async with self.session.get("{}/json".format(self.dev_url), timeout=timeout) as rp:
            tabs_map = {}
            for tab_json in await rp.json():
                if tab_json['type'] != 'page':  # pragma: no cover
                    continue

                if tab_json['id'] in self._tabs and self._tabs[tab_json['id']].status != Tab.status_stopped:
                    tabs_map[tab_json['id']] = self._tabs[tab_json['id']]
                else:
                    tabs_map[tab_json['id']] = Tab(**tab_json)

            self._tabs = tabs_map
            return list(self._tabs.values())

    async def activate_tab(self, tab_id, timeout=None):
        if isinstance(tab_id, Tab):
            tab_id = tab_id.id

        async with self.session.get("%s/json/activate/%s" % (self.dev_url, tab_id), timeout=timeout) as rp:
            return await rp.text()

    async def close_tab(self, tab_id, timeout=None):
        if isinstance(tab_id, Tab):
            tab_id = tab_id.id

        tab = self._tabs.pop(tab_id, None)
        if tab and tab.status == Tab.status_started:  # pragma: no cover
            tab.stop()

        async with self.session.get("%s/json/close/%s" % (self.dev_url, tab_id), timeout=timeout) as rp:
            return await rp.text()

    async def version(self, timeout=None):
        async with self.session.get("%s/json/version" % self.dev_url, json=True, timeout=timeout) as rp:
            return await rp.json()

    def __str__(self):
        return '<Browser %s>' % self.dev_url

    __repr__ = __str__

    def __del__(self):
        if not self.session.closed:
            self.session.close()



# -*- coding: utf-8 -*-

import pytest
import asyncio
import logging
import aiochrome


logging.basicConfig(level=logging.INFO)


async def close_all_tabs(browser):
    if len(await browser.list_tab()) == 0:
        return

    for tab in await browser.list_tab():
        await browser.close_tab(tab)

    await asyncio.sleep(1)
    assert len(await browser.list_tab()) == 0


@pytest.mark.asyncio
async def test_chome_version():
    browser = aiochrome.Browser()
    await close_all_tabs(browser)

    browser_version = await browser.version()
    assert isinstance(browser_version, dict)


@pytest.mark.asyncio
async def test_browser_list():
    browser = aiochrome.Browser()
    await close_all_tabs(browser)

    tabs = await browser.list_tab()
    assert len(tabs) == 0


@pytest.mark.asyncio
async def test_browser_new():
    browser = aiochrome.Browser()
    await close_all_tabs(browser)

    await browser.new_tab()
    tabs = await browser.list_tab()
    assert len(tabs) == 1


@pytest.mark.asyncio
async def test_browser_activate_tab():
    browser = aiochrome.Browser()
    await close_all_tabs(browser)

    tabs = []
    for i in range(10):
        tabs.append(await browser.new_tab())

    for tab in tabs:
        await browser.activate_tab(tab)


@pytest.mark.asyncio
async def test_browser_tabs_map():
    browser = aiochrome.Browser()
    await close_all_tabs(browser)

    tab = await browser.new_tab()
    assert tab in await browser.list_tab()

    await browser.close_tab(tab)
    assert tab not in await browser.list_tab()


@pytest.mark.asyncio
async def test_browser_new_10_tabs():
    browser = aiochrome.Browser()
    tabs = []
    for i in range(10):
        tabs.append(await browser.new_tab())

    await asyncio.sleep(1)
    assert len(await browser.list_tab()) == 10

    for tab in tabs:
        await browser.close_tab(tab.id)

    await asyncio.sleep(1)
    assert len(await browser.list_tab()) == 0


@pytest.mark.asyncio
async def test_browser_new_100_tabs():
    browser = aiochrome.Browser()
    tabs = []
    for i in range(100):
        tabs.append(await browser.new_tab())

    await asyncio.sleep(1)
    assert len(await browser.list_tab()) == 100

    for tab in tabs:
        await browser.close_tab(tab)

    await asyncio.sleep(1)
    assert len(await browser.list_tab()) == 0



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals


class AioChromeException(Exception):
    pass


class UserAbortException(AioChromeException):
    pass


class TabConnectionException(AioChromeException):
    pass


class CallMethodException(AioChromeException):
    pass


class TimeoutException(AioChromeException):
    pass


class RuntimeException(AioChromeException):
    pass


# -*- coding: utf-8 -*-

import os
import json
import asyncio
import logging
import warnings
import functools
import websockets

from .exceptions import *


__all__ = ["Tab"]


logger = logging.getLogger(__name__)


class GenericAttr:
    def __init__(self, name, tab):
        self.__dict__['name'] = name
        self.__dict__['tab'] = tab

    def __getattr__(self, item):
        method_name = "%s.%s" % (self.name, item)
        event_listener = self.tab.get_listener(method_name)

        if event_listener:
            return event_listener

        return functools.partial(self.tab.call_method, method_name)

    def __setattr__(self, key, value):
        self.tab.set_listener("%s.%s" % (self.name, key), value)


class Tab:
    status_initial = 'initial'
    status_started = 'started'
    status_stopped = 'stopped'

    def __init__(self, **kwargs):
        self.id = kwargs.get("id")
        self.type = kwargs.get("type")
        self.debug = os.getenv("DEBUG", False)
        self.loop = kwargs.get('loop', None) or asyncio.get_event_loop()

        self._websocket_url = kwargs.get("webSocketDebuggerUrl")
        self._kwargs = kwargs

        self._cur_id = 1000

        self._ws = None

        self._recv_task = None
        self._handle_event_task = None

        self._stopped = asyncio.Event()
        self._started = False
        self.status = self.status_initial

        self.event_handlers = {}
        self.method_results = {}
        self.event_queue = asyncio.Queue()

    async def _send(self, message, timeout=None):
        if 'id' not in message:
            self._cur_id += 1
            message['id'] = self._cur_id

        message_json = json.dumps(message)

        if self.debug:  # pragma: no cover
            print("SEND ► %s" % message_json)

        if not isinstance(timeout, (int, float)) or timeout > 1:
            q_timeout = 1
        else:
            q_timeout = timeout / 2.0

        try:
            queue = asyncio.Queue()
            self.method_results[message['id']] = queue

            # just raise the exception to user
            await self._ws.send(message_json)

            while not self._stopped.is_set():
                try:
                    if isinstance(timeout, (int, float)):
                        if timeout < q_timeout:
                            q_timeout = timeout

                        timeout -= q_timeout

                    return await asyncio.wait_for(queue.get(), q_timeout)
                except asyncio.TimeoutError:
                    if isinstance(timeout, (int, float)) and timeout <= 0:
                        raise TimeoutException("Calling %s timeout" % message['method'])

                    continue

            raise UserAbortException("User abort, call stop() when calling %s" % message['method'])
        finally:
            self.method_results.pop(message['id'])

    async def _recv_loop(self):
        while not self._stopped.is_set():
            try:
                message_json = await self._ws.recv()
                message = json.loads(message_json)
            except:
                continue

            if self.debug:  # pragma: no cover
                print('◀ RECV %s' % message_json)

            if "method" in message:
                await self.event_queue.put(message)

            elif "id" in message:
                if message["id"] in self.method_results:
                    await self.method_results[message['id']].put(message)
            else:  # pragma: no cover
                warnings.warn("unknown message: %s" % message)

    async def _handle_event_loop(self):
        while True:
            event = await self.event_queue.get()

            if event['method'] in self.event_handlers:
                try:
                    await self.event_handlers[event['method']](**event['params'])
                except Exception as e:
                    logger.error("callback %s exception" % event['method'], exc_info=True)

    def __getattr__(self, item):
        attr = GenericAttr(item, self)
        setattr(self, item, attr)
        return attr

    async def call_method(self, _method, *args, **kwargs):
        if not self._started:
            raise RuntimeException("Cannot call method before it is started")

        if args:
            raise CallMethodException("the params should be key=value format")

        if self._stopped.is_set():
            raise RuntimeException("Tab has been stopped")

        timeout = kwargs.pop("_timeout", None)
        result = await self._send({"method": _method, "params": kwargs}, timeout=timeout)
        if 'result' not in result and 'error' in result:
            warnings.warn("%s error: %s" % (_method, result['error']['message']))
            raise CallMethodException("calling method: %s error: %s" % (_method, result['error']['message']))

        return result['result']

    def set_listener(self, event, callback):
        if not callback:
            return self.event_handlers.pop(event, None)

        if not callable(callback):
            raise RuntimeException("callback should be callable")

        self.event_handlers[event] = callback
        return True

    def get_listener(self, event):
        return self.event_handlers.get(event, None)

    def del_all_listeners(self):
        self.event_handlers = {}
        return True

    async def start(self):
        if self._started:
            return False

        if not self._websocket_url:
            raise RuntimeException("Already has another client connect to this tab")

        self._started = True
        self.status = self.status_started
        self._stopped.clear()
        self._ws = await websockets.connect(self._websocket_url, loop=self.loop)

        self._recv_task = asyncio.ensure_future(self._recv_loop(), loop=self.loop)
        self._handle_event_task = asyncio.ensure_future(self._handle_event_loop(), loop=self.loop)

        return True

    async def stop(self):
        if self._stopped.is_set():
            return False

        if not self._started:
            raise RuntimeException("Tab is not running")

        self.status = self.status_stopped
        self._stopped.set()
        await self._ws.close()
        self._recv_task.cancel()
        self._handle_event_task.cancel()
        return True

    async def wait(self, timeout=None):
        if not self._started:
            raise RuntimeException("Tab is not running")

        if timeout:
            try:
                await asyncio.wait_for(self._stopped.wait(), timeout)
                return True
            except asyncio.TimeoutError:
                return False

        await self._recv_task
        await self._handle_event_task

    def __str__(self):
        return "<Tab [%s]>" % self.id

    __repr__ = __str__



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re
import ast
from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()


_version_re = re.compile(r'__version__\s+=\s+(.*)')


with open('aiochrome/__init__.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))


with open('requirements.txt', encoding='utf-8') as f:
    all_reqs = f.read().split('\n')


install_requires = [x.strip() for x in all_reqs if 'git+' not in x]
dependency_links = [x.strip().replace('git+', '') for x in all_reqs if x.startswith('git+')]


setup(
    name='aiochrome',
    version=version,
    description="",
    long_description=readme,
    author="fate0",
    author_email='fate0@fatezero.org',
    url='https://github.com/fate0/aiochrome',
    packages=find_packages(),
    package_dir={},
    entry_points={
        'console_scripts': [
            'aiochrome=aiochrome.cli:main'
        ]
    },
    include_package_data=True,
    install_requires=install_requires,
    dependency_links=dependency_links,
    license="Apache Software License 2.0",
    zip_safe=False,
    keywords='aiochrome',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)
